"""
Utility functions
"""

from __future__ import absolute_import, annotations

import logging
import time
from functools import wraps

logger = logging.getLogger(__name__)


def func_timer(f):
    """Decorator to record total execution time of a function to the configured logger using level DEBUG"""

    @wraps(f)
    def wrap(*args, **kw):
        """Wrapper function"""
        ts = time.time()
        result = f(*args, **kw)
        te = time.time()
        logger.debug(f"  [ {f.__name__}() ] took: {(te - ts):.6f} seconds")
        return result

    return wrap


def func_debug_logger(f):
    """Decorator for applying DEBUG logging to a function if enabled in the MBot class"""

    @wraps(f)
    def wrap(*args, **kw):
        """Wrapper function"""
        logger.debug(f"  [ function {f.__name__}() ] called with args: {args} and kwargs: {kw}")
        result = f(*args, **kw)
        return result

    return wrap



import pytest
from pytest_httpx import HTTPXMock

from mhanndalorian_bot.registry import Registry


@pytest.fixture(scope="package", autouse=True)
def registry_instance():
    """Fixture to create a Registry instance for testing."""
    return Registry(api_key="test_api_key", allycode="123456789", discord_id="123456789987654321")


def test_mock_fetch_player_valid_allycode(httpx_mock: HTTPXMock, registry_instance):
    """Test fetching a player with a valid allycode."""
    httpx_mock.add_response(json={"success": True}, status_code=200)
    response = registry_instance.fetch_player(allycode="123-456-789", hmac=True)
    assert response is not None
    assert isinstance(response, dict)


def test_fetch_player_invalid_allycode(httpx_mock: HTTPXMock, registry_instance):
    """Test fetching a player with an invalid allycode."""
    with pytest.raises(ValueError, match="Invalid allyCode"):
        response = registry_instance.fetch_player(allycode="invalid_allycode", hmac=True)


def test_register_player_valid_data(registry_instance):
    """Test registering a player with valid discord ID and allycode."""
    response = registry_instance.register_player(discord_id="123456789987654321", allycode="123-456-789", hmac=True)
    assert response is not None
    assert isinstance(response, dict)


def test_register_player_invalid_data(registry_instance):
    """Test registering a player with invalid data."""
    with pytest.raises(ValueError, match="Invalid"):
        response = registry_instance.register_player(discord_id="", allycode="invalid_allycode", hmac=True)


def test_verify_player_valid_data(registry_instance):
    """Test verifying a player with valid discord ID and allycode."""
    result = registry_instance.verify_player(discord_id="123456789987654321", allycode="123-456-789", primary=False, hmac=True)
    assert isinstance(result, bool)


def test_verify_player_invalid_data(registry_instance):
    """Test verifying a player with invalid data."""
    with pytest.raises(ValueError, match="Invalid"):
        result = registry_instance.verify_player(discord_id="", allycode="invalid_allycode", primary=False, hmac=True)



"""
Test script for MBot
"""

from __future__ import annotations, absolute_import
import asyncio
import logging
from pprint import pprint as pp

from mhanndalorian_bot import API, Registry, EndPoint

api = API(api_key="some_test_key", allycode="123456789")
api.sign(method="POST", endpoint=EndPoint.INVENTORY, payload={"payload": {"allyCode": "123456789"}})

allycode = '314927874'
allycode2 = '245866537'
api_key = 'usergeneratedoNtGtKG0YW6sLSj1Q12Q7hOCVhTAgvd0xMexdLK4qZZhnQbSO1'
discord_id = '344006512546545667'
discord_id2 = '730760587864178710'

# api = API(api_key, allycode)
# reg = Registry(api_key, allycode)

# resp = api.fetch_data(endpoint=EndPoint.INVENTORY,)

# if resp.status_code == 200:
#     resp_data = resp.json()
#     print(sorted(list(resp_data['inventory'].keys())))
# else:
#     print(f"{resp.status_code}: {resp.text}")

reg_resp = reg.fetch_player(allycode,)

if isinstance(reg_resp, dict) and 'allyCode' in reg_resp and 'discordId' in reg_resp:
    print("Registry fetch_user() verified.")

print("Registry register_user() test:")
reg_resp = reg.register_player(discord_id2, allycode2)
pp(reg_resp)

if isinstance(reg_resp, dict):
    print(reg_resp['unlockedPlayerPortrait'])
    print(reg_resp['unlockedPlayerTitle'])
    print("Registry register_user() verified.")

reg_resp = reg.verify_player(discord_id2, allycode2)
print(f"Registry verify_user(): {reg_resp}")


async def main():
    # reg = Registry(api_key, allycode)

    api = API(api_key, allycode)

    fetch_data_resp = await api.fetch_data_async(EndPoint.INVENTORY)

    if isinstance(fetch_data_resp, dict):
        if 'msg' in fetch_data_resp:
            # An unexpected error occurred
            print(f"An unexpected error occurred: {fetch_data_resp}")
        elif 'inventory' in fetch_data_resp:
            material: list = fetch_data_resp['inventory']['material']
            currency: list = fetch_data_resp['inventory']['currencyItem']
            equipment: list = fetch_data_resp['inventory']['equipment']
            unequipped_mods: list = fetch_data_resp['inventory']['unequippedMod']
        else:
            raise RuntimeError("Not sure what happened.")


    # Returns 'None' if player does not exist in the registry
    fetch_resp = await reg.fetch_player_async(allycode)

    if isinstance(fetch_resp, dict):
        if 'msg' in fetch_resp:
            # An unexpected error occurred
            print(f"An unexpected error occurred: {fetch_resp}")
        else:
            player_allycode = fetch_resp['allyCode']
            player_discord_id = fetch_resp['discordId']


    register_resp = await reg.register_player_async(discord_id, allycode2)
    pp(register_resp)
    pp(register_resp.text)
    if register_resp.status_code == 200:
        register_data = register_resp.json()
        pp(register_data)
        print(register_data['unlockedPlayerPortrait'])
        print(register_data['unlockedPlayerTitle'])
        print("Registry async register_user() verified.")

    # verify_resp = await reg.verify_player_async(discord_id, allycode2,)  # print(f"Registry async verify_user(): {verify_resp}")


if __name__ == '__main__':
    asyncio.run(main())



"""
Central shared configuration class
"""

from __future__ import annotations, absolute_import

import logging


class Config:
    """
    Provides a configuration template that ensures immutability and prevents instantiation
    or subclassing. It includes functionality to manage and retrieve a debug configuration
    state, ensuring robust control over debugging settings.

    The Config class is designed to act as a utility class, allowing only static interaction
    with its methods and attributes.

    Attributes:
        _debug (bool): Represents the debugging state of the application. Defaults to False.
    """
    _debug: bool = False

    logger = logging.getLogger(__name__)

    def __new__(cls, *args, **kwargs):
        raise TypeError(f"{cls.__name__} cannot be instantiated.")

    def __init_subclass__(cls, **kwargs):
        raise TypeError(f"{cls.__name__} cannot be subclassed.")

    @staticmethod
    def get_debug() -> bool:
        """Retrieve the current debug configuration state"""
        return Config._debug

    @staticmethod
    def set_debug(value: bool):
        """Set the debug configuration state"""
        if not isinstance(value, bool):
            raise ValueError("Debug value must be a boolean.")
        Config._debug = value



# tests/test_api.py
import pytest
from pytest_httpx import HTTPXMock

from mhanndalorian_bot.api import API
from mhanndalorian_bot.attrs import EndPoint

api_instance = API("mock_api_key", "123456789")


def test_mock_fetch_data_sync(httpx_mock: HTTPXMock):
    httpx_mock.add_response(json={"success": True}, status_code=200)
    response = api_instance.fetch_data(endpoint="mock_endpoint")
    assert response == {"success": True}


@pytest.mark.asyncio
async def test_mock_fetch_data_async(httpx_mock: HTTPXMock):
    httpx_mock.add_response(json={"success": True}, status_code=200)
    response = await api_instance.fetch_data_async(endpoint=EndPoint.TW)
    assert response == {"success": True}



# coding=utf-8
"""
Class definition for SWGOH MHanndalorian Bot player registry service
"""
from __future__ import absolute_import, annotations

import logging
from typing import Any, Dict, TYPE_CHECKING

if TYPE_CHECKING:
    import httpx

from mhanndalorian_bot.base import MBot
from mhanndalorian_bot.attrs import EndPoint
from mhanndalorian_bot.utils import func_timer


class Registry(MBot):
    """
    Container class for MBot module to facilitate interacting with Mhanndalorian Bot SWGOH player registry
    """

    logger = logging.getLogger(__name__)

    def __init__(self, api_key: str, allycode: str, discord_id: str, *, api_host: str = "https://mhanndalorianbot.work",
                 hmac: bool = True, debug: bool = False):
        super().__init__(api_key=api_key, allycode=allycode, discord_id=self.set_discord_id(discord_id),
                         api_host=api_host, hmac=hmac, debug=debug)

    @func_timer
    def fetch_player(self, allycode: str, *, hmac: bool = False) -> Dict[Any, Any]:
        """Return player data from the provided allycode

            Args
                allycode: Player allycode as a string.

            Keyword Args
                hmac: Boolean flag to indicate use of HMAC request signing.

            Returns
                Dictionary from JSON response, if found. Else None.
        """

        allycode = self.cleanse_allycode(allycode)

        payload = {'user': [allycode], 'endpoint': 'find'}
        endpoint = f"/api/{EndPoint.FETCH.value}"

        if hmac or self.hmac is True:
            self.sign(method='POST', endpoint=endpoint, payload=payload)

        resp: httpx.Response = self.client.post(endpoint, json=payload)

        if resp.status_code == 200:
            resp_data = resp.json()
            if isinstance(resp_data, list) and len(resp_data) == 1:
                return resp_data[0]
            else:
                return resp_data
        else:
            raise RuntimeError(f"Unexpected result: {resp.content.decode()}")

    @func_timer
    def register_player(self,
                        discord_id: str,
                        allycode: str, *, hmac: bool = False) -> Dict[str, Any]:
        """Register a player in the registry

            Args
                discord_id: Discord user ID as a string
                allycode: Player allycode as a string

            Keyword Args
                hmac: Boolean flag to indicate use of HMAC request signing.

            Returns
                Dict containing `unlockedPlayerPortrait` and `unlockedPlayerTitle` keys, if successful
        """

        allycode = self.cleanse_allycode(allycode)
        discord_id = self.cleanse_discord_id(discord_id)

        payload = dict(discordId=discord_id, method="registration", payload={"allyCode": allycode})
        endpoint = f"/api/{EndPoint.REGISTER.value}"

        if hmac or self.hmac is True:
            self.sign(method='POST', endpoint=endpoint, payload=payload)

        resp: httpx.Response = self.client.post(endpoint, json=payload)

        if resp.status_code == 200:
            return resp.json()
        else:
            raise RuntimeError(f"Unexpected result: {resp.content.decode()}")

    @func_timer
    def verify_player(self, discord_id: str, allycode: str, *, primary: bool = False, hmac: bool = False) -> bool:
        """Perform player portrait and title verification after register_player() has been called.

            Args
                discord_id: Discord user ID as a string.
                allycode: Player allycode as a string.

            Keyword Args
                primary: Boolean indicating whether this allycode should be used as the primary for the discord ID
                            in cases where multiple allycodes are registered to the same discord ID.
                hmac: Boolean flag to indicate use of HMAC request signing.

            Returns
                True if successful, False otherwise
        """

        allycode = self.cleanse_allycode(allycode)
        discord_id = self.cleanse_discord_id(discord_id)

        payload = dict(discordId=discord_id, method="verification", primary=primary, payload={"allyCode": allycode})
        endpoint = f"/api/{EndPoint.VERIFY.value}"

        if hmac or self.hmac is True:
            self.sign(method='POST', endpoint=endpoint, payload=payload)

        resp: httpx.Response = self.client.post(endpoint, json=payload)

        if resp.status_code == 200:
            resp_json = resp.json()
            if 'verified' in resp_json:
                return resp_json['verified']
        else:
            self.logger.error(f"Unexpected result: {resp.content.decode()}")

        return False

    # Async methods
    @func_timer
    async def fetch_player_async(self, allycode: str, *, hmac: bool = False) -> Dict[Any, Any]:
        """Return player data from the provided allycode

            Args
                allycode: Player allycode as a string.

            Keyword Args
                hmac: Boolean flag to indicate use of HMAC request signing.

            Returns
                Dictionary from JSON response, if found. Else None.
        """

        allycode = self.cleanse_allycode(allycode)

        payload = {'user': [allycode], 'endpoint': 'find'}
        endpoint = f"/api/{EndPoint.FETCH.value}"

        if hmac or self.hmac is True:
            self.sign(method='POST', endpoint=endpoint, payload=payload)

        result: httpx.Response = await self.aclient.post(endpoint, json=payload)

        if result.status_code == 200:
            resp_data = result.json()
            if isinstance(resp_data, list) and len(resp_data) == 1:
                return resp_data[0]
            else:
                return resp_data
        else:
            return {"msg": "Unexpected result", "reason": result.content.decode()}

    @func_timer
    async def register_player_async(self, discord_id: str, allycode: str, *, hmac: bool = False) -> Dict[Any, Any]:
        """Register a player in the registry

            Args
                discord_id: Discord user ID as a string.
                allycode: Player allycode as a string.

            Keyword Args
                hmac: Boolean flag to indicate use of HMAC request signing.

            Returns
                Dict containing `unlockedPlayerPortrait` and `unlockedPlayerTitle` keys, if successful.
        """

        allycode = self.cleanse_allycode(allycode)
        discord_id = self.cleanse_discord_id(discord_id)

        payload = dict(discordId=discord_id, method="registration", payload={"allyCode": allycode})
        endpoint = f"/api/{EndPoint.REGISTER.value}"

        if hmac or self.hmac is True:
            self.sign(method='POST', endpoint=endpoint, payload=payload)

        resp: httpx.Response = await self.aclient.post(endpoint, json=payload)

        if resp.status_code == 200:
            return resp.json()
        else:
            raise RuntimeError(f"Unexpected result: {resp.content.decode()}")

    @func_timer
    async def verify_player_async(self,
                                  discord_id: str,
                                  allycode: str, *,
                                  primary: bool = False, hmac: bool = False) -> bool:
        """Perform player portrait and title verification

            Args
                discord_id: Discord user ID as a string
                allycode: Player allycode as a string

            Keyword Args
                primary: Boolean indicating whether this allycode should be used as the primary for the discord ID
                            in cases where multiple allycodes are registered to the same discord ID
                hmac: Boolean flag to indicate use of HMAC request signing. Default: False.

            Returns
                True if successful, False otherwise
        """

        allycode = self.cleanse_allycode(allycode)
        discord_id = self.cleanse_discord_id(discord_id)

        payload = dict(discordId=discord_id, method="verification", primary=primary, payload={"allyCode": allycode})
        endpoint = f"/api/{EndPoint.VERIFY.value}"

        if hmac or self.hmac is True:
            self.sign(method='POST', endpoint=endpoint, payload=payload)

        resp: httpx.Response = await self.aclient.post(endpoint, json=payload)

        if resp.status_code == 200:
            resp_json = resp.json()
            if 'verified' in resp_json:
                return resp_json['verified']
        else:
            self.logger.error(f"Unexpected result: {resp.content.decode()}")

        return False



"""
Base class definition for core object for storing and sharing information for use by all modules
"""

from __future__ import absolute_import, annotations

import hashlib
import hmac
import logging
import time
from json import dumps
from typing import Any, Optional

import httpx
from sentinels import Sentinel

from mhanndalorian_bot.attrs import APIKey, AllyCode, EndPoint
from mhanndalorian_bot.utils import func_debug_logger, func_timer

# Define sentinels used in parameter checking
NotSet = Sentinel('NotSet')


class MBot:
    """Base class for MBot modules

    Args
        api_key: MHanndalorian Bot API key as a string
        allycode: Player allycode as a string
        discord_id: Discord user ID as a string. This is used to identify the source of requests made on behalf
                    of another player.

    Keyword Args
        api_host: Optional host URL for MHanndalorian Bot API, defaults to https://mhanndalorianbot.work/
        hmac: Boolean flag indicating whether the endpoints should use HMAC signature authentication, Default: True
        debug: Boolean flag indicating whether debug logging should be enabled, Default: False
    """

    api_host: str = "https://mhanndalorianbot.work"
    logger: logging.Logger = logging.getLogger(__name__)

    debug = False
    hmac = True
    api_key = APIKey()
    allycode = AllyCode()

    headers = {"Content-Type": "application/json"}
    payload = {"payload": {"allyCode": ""}}

    client: httpx.Client = httpx.Client(base_url=f"{api_host}", timeout=75, verify=False)
    aclient: httpx.AsyncClient = httpx.AsyncClient(base_url=f"{api_host}", timeout=75, verify=False)

    def __init__(self, api_key: str, allycode: str, discord_id: Optional[str] = None, *, api_host: Optional[str] = None,
                 hmac: Optional[bool] = True, debug: Optional[bool] = False):

        self.set_api_key(api_key)
        self.set_allycode(allycode)

        self.debug = debug

        if isinstance(api_host, str):
            self.set_api_host(api_host)

        if isinstance(hmac, bool):
            self.hmac = hmac

    @staticmethod
    @func_debug_logger
    def cleanse_allycode(allycode: str) -> str:
        """Remove any dashes from provided string and verify the result contains exactly 9 digits"""
        if not isinstance(allycode, str):
            raise ValueError(f"{allycode} must be a string, not type:{type(allycode)}")

        allycode = allycode.replace('-', '') if '-' in allycode else allycode

        if not allycode.isdigit() or len(allycode) != 9:
            raise ValueError(f"Invalid allyCode ({allycode}): Value must be exactly 9 numerical characters.")

        return allycode

    @staticmethod
    @func_debug_logger
    def cleanse_discord_id(discord_id: str) -> str:
        """Validate that discord ID is an 18 character string of only numerical digits"""
        if not isinstance(discord_id, str):
            raise ValueError(f"{discord_id} must be a string, not type: {type(discord_id)}")

        if not discord_id.isdigit() or len(discord_id) != 18:
            raise ValueError(f"Invalid Discord ID ({discord_id}): Value must be exactly 18 numerical characters.")

        return discord_id

    def get_api_key(self):
        """Return masked API key for logging purposes."""
        return f"{'*' * 4 + self.api_key[-4:]}"

    @func_debug_logger
    def set_api_key(self, api_key: str):
        """Set the api_key value for the container class and update relevant attributes (including headers)"""

        if not isinstance(api_key, str):
            raise ValueError("api_key must be a string")

        setattr(self, "api_key", api_key)

        self.headers["api-key"] = self.api_key
        self.client.headers = self.headers
        self.aclient.headers = self.headers

    @func_debug_logger
    def set_allycode(self, allycode: str):
        """Set the allycode value for the container class and update relevant attributes"""

        allycode = self.cleanse_allycode(allycode)

        setattr(self, "allycode", allycode)

        self.payload["payload"]["allyCode"] = allycode

    @func_debug_logger
    def set_discord_id(self, discord_id: str):
        """Set the discord_id value for the container class and update relevant attributes"""

        discord_id = self.cleanse_discord_id(discord_id)

        self.headers['x-discord-id'] = discord_id

    @func_debug_logger
    def set_api_host(self, api_host: str):
        """Set the api_host value for the container class and update relevant attributes"""

        if not isinstance(api_host, str):
            raise ValueError("api_host must be a string")

        setattr(self, "api_host", api_host)

        self.client.base_url = f"{api_host}"
        self.aclient.base_url = f"{api_host}"

    @func_debug_logger
    def set_client(self, **kwargs: Any):
        """Set the client values for the container class and update relevant attributes"""
        for key, value in kwargs.items():
            setattr(self.client, key, value)

    @func_timer
    @func_debug_logger
    def sign(self, method: str, endpoint: str | EndPoint, payload: dict[str, Any] | Sentinel = NotSet, *,
             timestamp: str = None, api_key: str = None) -> None:
        """Create HMAC signature for request

            Args
                method: HTTP method as a string
                endpoint: API endpoint path as a string or EndPoint enum instance
                payload: Dictionary containing API endpoint payload data.
                         This will be converted to a JSON string and hashed.
                         If no payload is provided, a default containing the currently set allyCode will be used.

            Keyword Args
                timestamp: Optional timestamp string to use instead of generating a new one. (primarily for testing)
                api_key: Optional API key to use instead of the one set in the container class. (primarily for testing)
        """
        # Removing 'api-key' header. This signals to the API that the request is HMAC signed
        if 'api-key' in self.headers:
            del self.headers['api-key']
            self.logger.debug("'api-key' header removed")

        # Get current unix timestamp in milliseconds
        if timestamp:
            req_time = timestamp
        else:
            req_time = str(int(time.time() * 1000))

        # Set the timestamp header for the API
        self.headers['x-timestamp'] = req_time
        self.logger.debug(f"'x-timestamp' header set to {self.headers['x-timestamp']}")

        # Create empty HMAC object using SHA256 algorithm
        if api_key:
            self.logger.debug(f"Using provided API key: [{self.get_api_key()}]")
            a_key = api_key.encode()
        else:
            self.logger.debug(f"Using API key from container class: [{self.get_api_key()}]")
            a_key = self.api_key.encode()
        hmac_obj = hmac.new(key=a_key, digestmod=hashlib.sha256)
        self.logger.debug(f"HMAC Hexdigest (base): {hmac_obj.hexdigest()}")

        # Add the timestamp to the HMAC object
        hmac_obj.update(req_time.encode())
        self.logger.debug(f"HMAC Hexdigest (timestamp): {hmac_obj.hexdigest()}")

        # Add the HTTP request method in all caps to the HMAC object
        hmac_obj.update(method.upper().encode())
        self.logger.debug(f"HMAC Hexdigest (HTTP method): {hmac_obj.hexdigest()}")

        # Add the API endpoint path to the HMAC object
        if isinstance(endpoint, EndPoint):
            endpoint = endpoint.value
        hmac_obj.update(endpoint.encode())
        self.logger.debug(f"HMAC Hexdigest (endpoint): {hmac_obj.hexdigest()}")

        # Create a serialized string of the payload object
        payload = self.payload if payload is NotSet else payload
        payload_str = dumps(payload, separators=(',', ':'))
        self.logger.debug(f"Payload string: {payload_str}")

        # Generate MD5 hash of the payload string
        payload_hash_digest = hashlib.md5(payload_str.encode()).hexdigest()
        self.logger.debug(f"Payload hash digest: {payload_hash_digest}")

        # Add the payload MD5 hash to the HMAC object
        hmac_obj.update(payload_hash_digest.encode())
        self.logger.debug(f"HMAC Hexdigest (payload): {hmac_obj.hexdigest()}")

        # Add the HMAC signature to the HTTP headers
        self.headers['Authorization'] = hmac_obj.hexdigest()
        self.client.headers = self.headers
        self.aclient.headers = self.headers
        self.logger.debug(f"HTTP client headers updated with HMAC signature: {self.client.headers}")
